from pathlib import Path

from os import SEEK_SET, SEEK_END
import numpy as np

from ccvtools.rawio.ccv_header import CamCommandoHeader
from ccvtools.rawio.utils import unpack_bits_16_10

from construct import Int32ul, Float64l


class CamCommandoVideo:
    def __init__(self, filename, rw=False):
        self.filename = Path(filename)
        self._fh = None  # Lazy initialization, only access via get_file_handle
        self._rw = rw
        self._header_written = False  # The header has been written and may not change in length anymore
        self._header_updated = False  # The header construct has been updated and needs to be written to the file

        if self.filename.is_file():
            self.read_header()
            self._header_written = True
        else:
            self.header = CamCommandoHeader.parse(Int32ul.build(106) + 102 * b'\00')

    def __del__(self):
        self.close()

    def __len__(self):
        return self.header.frame_count

    def close(self):
        if self._header_updated:
            self.write_header()
        if self._fh is not None:
            self._fh.close()

    def get_file_handle(self):
        if self._fh is None:
            if self._rw:
                file_mode = 'r+b'
            else:
                file_mode = 'rb'
            if not self.filename.is_file():
                with open(self.filename, 'wb'):
                    pass

            self._fh = open(self.filename, file_mode)
        return self._fh

    def read_header_length(self, file_name=None):
        if file_name is None:
            fh = self.get_file_handle()
            fh.seek(0)
            header_length = int.from_bytes(fh.read(4), 'little')
        else:
            with open(file_name, 'rb') as fh:
                header_length = int.from_bytes(fh.read(4), 'little')
        return header_length

    def read_header(self):
        header_length = self.read_header_length()
        assert header_length < 1024, "Header is too long, you are probably not using a CCV file"
        fh = self.get_file_handle()
        fh.seek(0)
        self.header = CamCommandoHeader.parse(fh.read(header_length))

    def copy_header_from_file(self, file_name, zero_frame_count=True):
        header_length = self.read_header_length(file_name)
        assert header_length < 1024, "Header is too long, you are probably not using a CCV file"
        with open(file_name, 'rb') as fh:
            self.header = CamCommandoHeader.parse(fh.read(header_length))
        if zero_frame_count:
            self.header.frame_count = 0
        self._header_updated = True

    def write_header(self):
        if self._header_updated:
            header_bytes = CamCommandoHeader.build(self.header)
            if self._header_written:
                header_length = self.read_header_length()
                assert header_length == len(header_bytes), (
                    "Currently, we cannot write a header with a length different "
                    "from the original one.")
            fh = self.get_file_handle()
            fh.seek(0)
            fh.write(header_bytes)
            self._header_written = True
            self._header_updated = False

    def read_image(self, index=None, return_time=True):
        fh = self.get_file_handle()

        # Check if requested index is in range
        if index is not None:
            if index >= self.header.frame_count:
                raise IndexError(f"Image index {index} > {self.header.frame_count}")

            # Seek to request frame in file
            offset = self.header.header_size + self.header.frame_bytes_on_disk * index
            fh.seek(offset, SEEK_SET)

        # Read frame from file
        dimension = (self.header.height, self.header.width)
        if self.header.packed:
            # Packed files with 10 bits per pixel
            rawdata = np.fromfile(fh, np.uint8,
                                  int(np.ceil(dimension[0] * dimension[1] * self.header.bits_per_pixel / 8)))
            frame = unpack_bits_16_10(rawdata).reshape(dimension)  # type: np.uint16

        else:
            # Unpacked files with 8 bits per pixel
            frame = np.fromfile(fh, np.uint8, dimension[0] * dimension[1]).reshape(dimension)

        index = Int32ul.parse_stream(fh)
        timestamp = Float64l.parse_stream(fh)
        rest = fh.read(9)

        if return_time:
            # Read in additional fields

            return frame, {"index": index, "timestamp": timestamp, 'rest': rest}
        else:
            return frame

    def append_image(self, image: np.ndarray, index=None, meta=None):
        assert not self.header.packed, "Packed images currently not supported"
        assert image.dtype == np.uint8, "Only np.uint8 images are supported"

        if not self._header_written:
            self.write_header()

        if meta is None:
            meta = dict()
        meta_default = {
            "index": index,
            "timestamp": 0,
            "rest": 9 * b'\0',  # No idea what this is. Once found out, will be replaced
        }
        for key in meta_default:
            if key not in meta:
                meta[key] = meta_default[key]

        if index == -1:
            meta["index"] = self.header.frame_count
        elif index is not None:
            meta["index"] = index

        fh = self.get_file_handle()
        fh.seek(0, SEEK_END)
        pos = fh.tell()
        fh.write(image.tobytes())
        fh.write(Int32ul.build(meta["index"]))
        fh.write(Float64l.build(meta["timestamp"]))
        fh.write(meta["rest"])
        bytes_missing = self.header.frame_bytes_on_disk - (fh.tell() - pos)
        assert bytes_missing == 0, (f"Written {bytes_missing} bytes too few. File is corrupt now!")

        self.header.frame_count += 1
        self._header_updated = True



import numpy as np


def unpack_bits_16_10(input_array: np.ndarray):
  
    """Unpacks data into uint16 values from uint8 packed values. Length of each sample(or pixel) is 10 bits."""
    # Written based on bboanalysis_m/datamanipulation/unpack_bits.cpp and C++ funtion 'unpack_bits_16_10' in 'bboanalysis_m/camera cpp code/dgpackbits.h'.

    nbytes_in = input_array.size
    nsamples_in = int(np.floor(nbytes_in * 8.0 / 10.0))

    bytes_per_sample_unpacked = 2  # int(np.ceil(10.0 / 8.0))
    nbytes_out = bytes_per_sample_unpacked * nsamples_in

    nblocks = nsamples_in / 4.0
    nblocks_floor = int(np.floor(nblocks))

    # The final output array
    output_array = np.zeros(nbytes_out, dtype='uint8')

    input_array_1 = input_array[:5 * nblocks_floor].reshape(nblocks_floor, 5)
    output_array_1 = np.zeros((nblocks_floor, 8), dtype='uint8')

    # Process the data in blocks of 40 bits of data = 4 samples = 8 unpacked bytes = 5 packed bytes
    output_array_1[:, 0] = input_array_1[:, 0]
    output_array_1[:, 1] = input_array_1[:, 1] & 3
    output_array_1[:, 2] = (input_array_1[:, 1] >> 2) | (
            (input_array_1[:, 2] & 3) << 6)
    output_array_1[:, 3] = (input_array_1[:, 2] >> 2) & 3
    output_array_1[:, 4] = (input_array_1[:, 2] >> 4) | \
                                        (input_array_1[:, 3] << 4)
    output_array_1[:, 5] = (input_array_1[:, 3] >> 4) & 3
    output_array_1[:, 6] = (input_array_1[:, 3] >> 6) | \
                                        (input_array_1[:, 4] << 2)
    output_array_1[:, 7] = input_array_1[:, 4] >> 6

    output_array[:8 * nblocks_floor] = output_array_1.flatten()

    # Unpacking the remaining samples after the blocks
    remaining_samples = nsamples_in - (4 * nblocks_floor)

    if remaining_samples != 0:
        byte_offset_in = int(5 * nblocks_floor)
        byte_offset_out = int(8 * nblocks_floor)
        
        output_array[byte_offset_out] = input_array[byte_offset_in]
        output_array[byte_offset_out + 1] = input_array[byte_offset_in + 1] & 3
        
        if remaining_samples > 1:
            output_array[byte_offset_out + 2] = (input_array[byte_offset_in + 1] >> 2) | \
                                                ((input_array[byte_offset_in + 2] & 3) << 6)
            output_array[byte_offset_out + 3] = (input_array[byte_offset_in + 2] >> 2) & 3
            
            if remaining_samples > 2:
                output_array[byte_offset_out + 4] = (input_array[byte_offset_in + 2] >> 4) | \
                                                    (input_array[byte_offset_in + 3] << 4)
                output_array[byte_offset_out + 5] = (input_array[byte_offset_in + 3] >> 4) & 3
    
    # Two unpacked bytes per pixel are processed into uint16 value.
    output_array = output_array.astype('uint16').reshape(-1, 2)
    output_array = (output_array[:, 1] << 8) | output_array[:, 0]
    
    return output_array



from construct import this, Struct, Bytes, Const, If, Int64ul, Int32ul, Float64l, Byte

GreenbergString = Struct(
    "length" / Int32ul,  # > 1 < 1e3
    "data" / Bytes(this.length),
    Const(0, Byte)
)

CamCommandoHeader = Struct(
    "header_size" / Int32ul,  # < 1e5
    "camera_type" / GreenbergString,  # b'basler', b'aptina'
    "header_version" / Float64l,  # > 0 < 100 > 0.13
    "image_type" / GreenbergString,

    "bytes_per_pixel" / Int32ul,
    "bits_per_pixel" / Int32ul,
    "frame_bytes_on_disk" / Int32ul,

    "width" / Int32ul,
    "height" / Int32ul,
    "frame_rate" / Float64l,

    "packed" / If(this.header_version >= 0.12, Byte),

    "frame_count" / Int32ul,

    "sensor" / If(this.header_version >= 0.12, Struct(
        "offset" / Int32ul[2],
        "size" / Int32ul[2],
        "clock" / Int64ul,
        "exposure" / Float64l,
        "gain" / Float64l
    ))
    # ToDo: Check that we are not exceeding header size here!
)


# (c) 2021 Kay-Michael Voit <kay@voits.net>

import os
import numpy as np
import imageio
from ccvtools import rawio  # noqa
import random
import string
import mutagen
import json
from PIL import Image


def fill(full_times_file, ccv_file, ccv_times_file):
    ccv_out_file = ccv_file[0:-4] + '_filled' + ccv_file[-4:]

    ccv_times = np.loadtxt(ccv_times_file)
    full_times = np.loadtxt(full_times_file)

    assert ccv_times[0] > full_times[0] - 0.001
    doesnotexist = np.zeros(full_times.size, dtype=np.bool)
    ccv_times_idx = 0
    for i, t in enumerate(full_times):
        if len(ccv_times) <= ccv_times_idx:
            doesnotexist[i:] = True
            break
        if not np.abs(ccv_times[ccv_times_idx] - t) < 0.001:
            doesnotexist[i] = True
        else:
            ccv_times_idx = ccv_times_idx + 1

    reader = imageio.get_reader(ccv_file)

    camera_type_len = reader.header.camera_type.length + 5
    image_type_len = reader.header.image_type.length + 5
    frcount_offset = int((8 * 32 + 64 + 8) / 8 + camera_type_len + image_type_len)

    blackframe = np.zeros(reader.header.frame_bytes_on_disk, dtype=np.uint8)

    with open(ccv_file, 'rb') as sourcefile:
        with open(ccv_out_file, 'wb') as targetfile:
            targetfile.write(sourcefile.read(frcount_offset))
            targetfile.write(np.uint32(len(full_times)))
            sourcefile.seek(frcount_offset + 4, os.SEEK_SET)
            targetfile.write(sourcefile.read(reader.header.header_size - 4 - frcount_offset))

            for writeblack in doesnotexist:
                if writeblack:
                    targetfile.write(blackframe)
                else:
                    targetfile.write(sourcefile.read(reader.header.frame_bytes_on_disk))


def truncate(ccv_file, idx_range, ccv_out_file=None):
    reader = imageio.get_reader(ccv_file)
    camera_type_len = reader.header.camera_type.length + 5
    image_type_len = reader.header.image_type.length + 5
    frcount_offset = int((8 * 32 + 64 + 8) / 8 + camera_type_len + image_type_len)

    if ccv_out_file is None:
        ccv_tmp_file = f'.{ccv_file[0:-4]}_{"".join(random.choices(string.ascii_uppercase + string.digits, k=5))}_truncate.tmp'
    else:
        ccv_tmp_file = f'.{ccv_out_file[0:-4]}_{"".join(random.choices(string.ascii_uppercase + string.digits, k=5))}_truncate.tmp'

    with open(ccv_file, 'rb') as sourcefile:
        with open(ccv_tmp_file, 'wb') as targetfile:
            # Copy header
            targetfile.write(sourcefile.read(reader.header.header_size))

            # Write frame number to header
            targetfile.seek(frcount_offset, os.SEEK_SET)
            targetfile.write(np.uint32(len(idx_range)))
            targetfile.seek(reader.header.header_size, os.SEEK_SET)

            # Copy desired frames
            prev_fr_idx = None
            for (i, fr_idx) in enumerate(idx_range):
                # Save first frame from iterator
                if i == 0:
                    start_frame = fr_idx

                # Set point in case iterator is not consecutive
                if prev_fr_idx is None or not fr_idx - prev_fr_idx == 1:
                    sourcefile.seek(reader.header.header_size + reader.header.frame_bytes_on_disk * fr_idx, os.SEEK_SET)

                targetfile.write(sourcefile.read(reader.header.frame_bytes_on_disk))
                prev_fr_idx = fr_idx

            # Save last frame from iterator
            end_frame = fr_idx

    # Rename tmp file to final file name
    if ccv_out_file is None:
        ccv_out_file = "{ccv_file[0:-4]}_truncated_{start_frame}-{end_frame}.ccv".format(ccv_file=ccv_file,
                                                                                         start_frame=start_frame,
                                                                                         end_frame=end_frame)
    os.rename(ccv_tmp_file, ccv_out_file)


def convert(ccv_file, video_file, idx_range, fps=25, codec="libx264", quality=10, min_contrast=0, max_contrast=None,
            out_type=np.uint8, resize=None):
    print(ccv_file)
    print(video_file)
    reader = imageio.get_reader(ccv_file)
    writer = imageio.get_writer(video_file, fps=fps, codec=codec, quality=quality)

    if idx_range is None:
        idx_range = range(reader.header.frame_count)

    prev_fr_idx = None
    indices = np.zeros(shape=len(idx_range), dtype=np.int64)
    timestamps = np.zeros(shape=len(idx_range), dtype=np.float64)
    for (i, fr_idx) in enumerate(idx_range):
        # Set point in case iterator is not consecutive
        if prev_fr_idx is None or not fr_idx - prev_fr_idx == 1:
            im = reader.get_data(fr_idx)
        else:
            im = reader.get_next_data()

        if resize is not None:
            if isinstance(resize, str):
                resize = [int(s) for s in resize.split(",")]
            elif isinstance(resize, int):
                im_shape = np.asarray(im.shape)
                im_shape[0:2] = im_shape[0:2]/resize
                resize = im_shape

            im = Image.fromarray(im).resize(resize)

        meta = reader.get_meta_data(fr_idx)
        indices[i] = meta["index"]
        timestamps[i] = meta["timestamp"]

        # Get max value of movie if not specified
        if i == 0 and max_contrast is None:
            max_contrast = np.iinfo(np.asarray(im).dtype).max

        # Adjust contrast
        np.clip(im, min_contrast, max_contrast, out=im)
        im -= min_contrast
        if np.iinfo(out_type).max != max_contrast - min_contrast:
            im = np.uint64(im)
            im = im * np.iinfo(out_type).max / (max_contrast - min_contrast)
        writer.append_data(out_type(im))
        prev_fr_idx = fr_idx

    writer.close()

    headerdict = dict(reader.header)
    del headerdict['_io']
    headerdict['camera_type'] = headerdict['camera_type'].data.decode("ascii")
    headerdict['image_type'] = headerdict['image_type'].data.decode("ascii")
    headerdict['sensor'] = {'offset': list(headerdict['sensor'].offset),
                            'size': list(headerdict['sensor'].size),
                            'clock': headerdict['sensor'].clock,
                            'exposure': headerdict['sensor'].exposure,
                            'gain': headerdict['sensor'].gain
                            }
    headerdict['indices'] = [int(val) for val in indices]
    print(timestamps)
    headerdict['timestamps'] = [float(val) for val in timestamps]
    print(headerdict['timestamps'])

    if len(headerdict['sensor']['offset']) == 1:
        headerdict['sensor']['offset'].append(headerdict['sensor']['offset'][0])
    if len(headerdict['sensor']['size']) == 1:
        headerdict['sensor']['size'].append(headerdict['sensor']['size'][0])

    with open(video_file, 'r+b') as file:
        media_file = mutagen.File(file, easy=True)
        media_file['description'] = json.dumps(headerdict)
        media_file.save(file)



# (c) 2019 Florian Franzen <Florian.Franzen@gmail.com>

from imageio import formats

from .CamCommandoFormat import CamCommandoFormat

from .Filters import FILTERS


# Initiate and register CamCommandoFormat
_ccv_format_instance = CamCommandoFormat(
    "CamCommandoVideo",
    "Reads in BBO CamCommando videos",
    ".ccv",
    "I",  # Video only
)
formats.add_format(_ccv_format_instance)



import argparse
import os,sys
import imageio

from . import ccv
from . import rawio

def main():
    parser = argparse.ArgumentParser(description="Detect positions in multitrack file")
    parser.add_argument('CCV_FILE', type=str, help="CCV File to work on")
    parser.add_argument('--action', '-a', required=True, nargs=1, choices=("convert",))
    parser.add_argument('--outfile', '-o', required=False, nargs=1, type=str, help="Appropriate extension will be added if not given")
    parser.add_argument('--fps', required=False, default=[25], nargs=1, type=int)
    parser.add_argument('--codec', required=False, default=['libx264'], nargs=1, type=str)
    parser.add_argument('--quality', required=False, default=[10], nargs=1, type=int)
    parser.add_argument('--idxrange', required=False, nargs=2, type=int, help="Index range. Attention! Python indexing!")
    parser.add_argument('--maxcontrast', required=False, nargs=1, type=int, help="Maximum brightness", default=None)

    args = parser.parse_args()
    
    if not os.access(args.CCV_FILE, os.R_OK):
        print(f"File {args.CCV_FILE} does not exist or isn't readable", file=sys.stderr)
        sys.exit(1)
    
    reader = imageio.get_reader(args.CCV_FILE);
    
    if args.idxrange is None:
        idx_range = range(reader.header.frame_count)
    else:
        idx_range = range(args.idxrange[0],args.idxrange[1])
    
    if args.action[0]=='convert':
        print('Converting ...')
        if args.outfile is None:
            video_file = args.CCV_FILE
        else:
            video_file = args.outfile[0]

        if not video_file[-4:]=='.mp4': video_file = video_file + '.mp4'
        print(video_file)
        ccv.convert(args.CCV_FILE, video_file, idx_range, fps=args.fps[0], codec=args.codec[0], quality=args.quality[0],
                    min_contrast=0, max_contrast=args.maxcontrast[0])
if __name__ == "__main__":
    main()
